import { global } from "../env/global";
const {XMLHttpRequest, Blob, TextDecoder} = global;
const REGEXP_HEADER_SEPARATOR = /[\r\n]+/;
const fetchLikeRequest = (url, {method = "GET", headers: requestHeaders, body, timeout = 0, credentials, onUploadProgress, onDownloadProgress} = {}) => new Promise((resolve, reject) => {
  const getError = (message, status) => Object.assign(new Error(message), {
    status,
    url,
    method
  });
  const request = new XMLHttpRequest();
  request.onerror = () => reject(getError("NETWORK_ERROR", -1));
  request.ontimeout = () => reject(getError("NETWORK_TIMEOUT", -1));
  request.onreadystatechange = () => {
    const {readyState, status} = request;
    if (2 !== readyState) return;
    if (0 === status) {
      request.abort();
      return reject(getError("HEADER_STATUS_ERROR", -1));
    }
    const responseHeaders = request.getAllResponseHeaders().split(REGEXP_HEADER_SEPARATOR).reduce((o, rawHeader) => {
      const [key, ...valueList] = rawHeader.split(":");
      valueList.length && (o[key.trim().toLowerCase()] = valueList.join(":").trim());
      return o;
    }, {});
    resolve({
      status,
      ok: status >= 200 && status < 300,
      headers: responseHeaders,
      ...wrapPayload(request, getError)
    });
  };
  onUploadProgress && request.upload && (request.upload.onprogress = onUploadProgress);
  onDownloadProgress && (request.onprogress = onDownloadProgress);
  request.open(method, url);
  requestHeaders && Object.entries(requestHeaders).forEach(([key, value]) => request.setRequestHeader(key, value));
  request.responseType = "arraybuffer";
  request.timeout = timeout || 0;
  request.withCredentials = "include" === credentials;
  request.send(body || null);
});
const wrapPayload = (request, getError) => {
  let isKeep;
  let isDropped;
  setTimeout(() => {
    if (!isKeep) {
      isDropped = !0;
      request.abort();
    }
  });
  const arrayBuffer = () => {
    if (isKeep) throw getError("PAYLOAD_ALREADY_USED", -1);
    if (isDropped) throw getError("PAYLOAD_ALREADY_DROPPED", -1);
    isKeep = !0;
    return new Promise((resolve, reject) => {
      request.onerror = () => reject(getError("PAYLOAD_ERROR", -1));
      request.ontimeout = () => reject(getError("PAYLOAD_TIMEOUT", -1));
      request.onreadystatechange = () => {
        if (4 === request.readyState) {
          if (0 === request.status) return reject(getError("PAYLOAD_STATUS_ERROR", -1));
          resolve(request.response);
        }
      };
    });
  };
  const text = () => arrayBuffer().then(toText);
  return {
    arrayBuffer,
    blob: () => arrayBuffer().then(toBlob),
    text,
    json: () => text().then(parseJSON)
  };
};
const toBlob = arrayBuffer => new Blob([ arrayBuffer ]);
const toText = arrayBuffer => new TextDecoder().decode(arrayBuffer);
const parseJSON = text => JSON.parse(text);
export { fetchLikeRequest };