import { createHub } from "../module/Event";
import { createDoublyLinkedList, createNode } from "./LinkedList";
const createCache = (key, value, size, expireAt) => ({
  ...createNode(value),
  key,
  size,
  expireAt
});
const createCacheMap = ({valueSizeSumMax, valueSizeSingleMax = Math.max(.05 * valueSizeSumMax, 1), eventHub = createHub()}) => {
  const hasEventHub = Boolean(eventHub);
  const {clear: clearEventHub, subscribe, unsubscribe, send} = hasEventHub ? eventHub : {};
  const map = new Map();
  const linkedList = createDoublyLinkedList();
  let valueSizeSum = 0;
  const cacheAdd = cache => {
    map.set(cache.key, cache);
    linkedList.unshift(cache);
    valueSizeSum += cache.size;
    hasEventHub && send({
      type: "add",
      key: cache.key,
      payload: cache.value
    });
  };
  const cacheDelete = cache => {
    map.delete(cache.key);
    linkedList.remove(cache);
    valueSizeSum -= cache.size;
    hasEventHub && send({
      type: "delete",
      key: cache.key,
      payload: cache.value
    });
  };
  return {
    hasEventHub,
    clearEventHub,
    subscribe,
    unsubscribe,
    clear: () => map.forEach(cacheDelete),
    getSize: linkedList.getLength,
    getValueSizeSum: () => valueSizeSum,
    set: (key, value, size = 1, expireAt = Date.now() + 6e4) => {
      const prevCache = map.get(key);
      prevCache && cacheDelete(prevCache);
      if (!(size > valueSizeSingleMax)) {
        for (;size + valueSizeSum > valueSizeSumMax; ) cacheDelete(linkedList.getTail().prev);
        cacheAdd(createCache(key, value, size, expireAt));
      }
    },
    get: (key, time = Date.now()) => {
      const cache = map.get(key);
      if (cache) {
        if (cache.expireAt <= time) return cacheDelete(cache);
        linkedList.moveToFirst(cache);
        return cache.value;
      }
    },
    touch: (key, expireAt = Date.now() + 6e4) => {
      const cache = map.get(key);
      if (cache) {
        cache.expireAt = expireAt;
        linkedList.moveToFirst(cache);
        return cache.value;
      }
    },
    delete: key => {
      const cache = map.get(key);
      cache && cacheDelete(cache);
      return cache && cache.value;
    },
    saveCacheList: () => {
      const cacheList = [];
      linkedList.forEachReverse(({key, value, size, expireAt}) => cacheList.push({
        key,
        value,
        size,
        expireAt
      }));
      return cacheList;
    },
    loadCacheList: (cacheList, time = Date.now()) => cacheList.forEach(({key, value, size, expireAt}) => {
      if (expireAt <= time) return;
      const cache = map.get(key);
      if (!(cache && cache.expireAt >= expireAt)) {
        cache && cacheDelete(cache);
        cacheAdd(createCache(key, value, size, expireAt));
      }
    })
  };
};
export { createCache, createCacheMap };