import { clock, setTimeoutAsync } from "./time";
import { rethrowError } from "./error";
import { isPromiseAlike } from "./check";
const debounce = (func, wait = 250, isLeadingEdge = !1) => {
  let timeoutToken = null;
  return (...args) => {
    const isCallNow = isLeadingEdge && null === timeoutToken;
    clearTimeout(timeoutToken);
    timeoutToken = setTimeout(() => {
      timeoutToken = null;
      !isLeadingEdge && func.apply(null, args);
    }, wait);
    isCallNow && func.apply(null, args);
  };
};
const throttle = (func, wait = 250, isLeadingEdge = !1) => {
  let timeoutToken = null;
  return (...args) => {
    if (timeoutToken) return;
    const isCallNow = isLeadingEdge && null === timeoutToken;
    timeoutToken = setTimeout(() => {
      timeoutToken = null;
      !isLeadingEdge && func.apply(null, args);
    }, wait);
    isCallNow && func.apply(null, args);
  };
};
const lossyAsync = (func, onError = rethrowError) => {
  let runningPromise;
  const onResolve = () => {
    runningPromise = void 0;
  };
  const onReject = error => {
    runningPromise = void 0;
    onError(error);
  };
  return {
    trigger: (...args) => {
      if (!runningPromise) try {
        const result = func.apply(null, args);
        isPromiseAlike(result) ? runningPromise = result.then(onResolve, onReject) : onResolve();
      } catch (error) {
        onReject(error);
      }
    },
    getRunningPromise: () => runningPromise
  };
};
const withDelayArgvQueue = (func, delayWrapper = debounce, ...args) => {
  let argvQueue = [];
  const delayFunc = delayWrapper(() => {
    const currentArgvQueue = argvQueue;
    argvQueue = [];
    func(currentArgvQueue);
  }, ...args);
  return (...argv) => {
    argvQueue.push(argv);
    delayFunc();
  };
};
const withRepeat = (func, count = 0) => {
  let looped = 0;
  for (;count > looped; ) {
    func(looped, count);
    looped++;
  }
};
const withRepeatAsync = async (func, count = 0, wait = 0) => {
  let looped = 0;
  for (;count > looped; ) {
    const startTime = clock();
    await func(looped, count);
    looped++;
    const remainingTime = wait - (clock() - startTime);
    remainingTime > 0 && await setTimeoutAsync(remainingTime);
  }
};
const withRetry = (func, maxRetry = 1 / 0) => {
  let failed = 0;
  for (;;) try {
    return func(failed, maxRetry);
  } catch (error) {
    if (maxRetry < ++failed) throw error;
  }
};
const withRetryAsync = async (func, maxRetry = 1 / 0, wait = 0) => {
  let failed = 0;
  for (;;) {
    const startTime = clock();
    try {
      return await func(failed, maxRetry);
    } catch (error) {
      if (maxRetry < ++failed) throw error;
      const remainingTime = wait - (clock() - startTime);
      remainingTime > 0 && await setTimeoutAsync(remainingTime);
    }
  }
};
const withTimeoutAsync = (func, timeout) => withTimeoutPromise(func(), timeout);
const withTimeoutPromise = (promise, timeout) => {
  let timeoutToken = null;
  return Promise.race([ promise, new Promise((resolve, reject) => {
    timeoutToken = setTimeout(() => reject(new Error(`timeout after: ${timeout}`)), timeout);
  }) ]).then(result => {
    clearTimeout(timeoutToken);
    return result;
  }, error => {
    clearTimeout(timeoutToken);
    throw error;
  });
};
const createInsideOutPromise = () => {
  let promiseResolve, promiseReject;
  return {
    promise: new Promise((resolve, reject) => {
      promiseResolve = resolve;
      promiseReject = reject;
    }),
    resolve: value => {
      const resolve = promiseResolve;
      promiseResolve = promiseReject = void 0;
      resolve && resolve(value);
    },
    reject: error => {
      const reject = promiseReject;
      promiseResolve = promiseReject = void 0;
      reject && reject(error);
    }
  };
};
export { debounce, throttle, lossyAsync, withDelayArgvQueue, withRepeat, withRepeatAsync, withRetry, withRetryAsync, withTimeoutAsync, withTimeoutPromise, createInsideOutPromise };