import { join as joinPath } from "path";
import { catchAsync } from "../../common/error";
import { lossyAsync } from "../../common/function";
import { tryParseJSONObject } from "../../common/data/function";
import { createStateStore } from "../../common/immutable/StateStore";
import { readFileAsync, writeFileAsync, unlinkAsync, createReadlineFromFileAsync } from "../file/function";
import { getDirectorySubInfoList } from "../file/Directory";
import { createLogger } from "./Logger";
const INITIAL_STATE = {
  id: 0
};
const INITIAL_FACT_INFO = {
  factId: 0,
  factState: INITIAL_STATE,
  factCacheFile: "",
  factLogFile: ""
};
const createFactDatabase = async ({initialFactInfo, initialState = INITIAL_STATE, applyFact = (state, fact) => ({
  ...state,
  ...fact
}), encodeFact = JSON.stringify, decodeFact = JSON.parse, pathFactDirectory, nameFactLogFile = "factLog", nameFactCacheFile = "factCache", onError, ...extraOption}) => {
  void 0 === initialFactInfo && (initialFactInfo = await tryLoadFactInfo({
    ...INITIAL_FACT_INFO,
    factState: initialState
  }, {
    applyFact,
    decodeFact,
    pathFactDirectory,
    nameFactLogFile,
    nameFactCacheFile
  }));
  let isActive = !0;
  let factId = initialFactInfo.factId || 0;
  let prevFactCacheFile = initialFactInfo.factCacheFile || "";
  const {getState, setState, subscribe, unsubscribe} = createStateStore(initialFactInfo.factState);
  const factLogger = await createLogger({
    ...extraOption,
    pathLogDirectory: pathFactDirectory,
    getLogFileName: () => `${nameFactLogFile}.${factId + 1}.log`,
    onError
  });
  const {trigger: saveFactCache, getRunningPromise: getSaveFactCachePromise} = lossyAsync(async () => {
    const factCacheFile = joinPath(pathFactDirectory, `${nameFactCacheFile}.${factId}.json`);
    if (prevFactCacheFile === factCacheFile) return;
    const fileContent = JSON.stringify({
      factId,
      factState: getState()
    });
    await writeFileAsync(factCacheFile, fileContent);
    prevFactCacheFile && await unlinkAsync(prevFactCacheFile).catch(error => {});
    prevFactCacheFile = factCacheFile;
  }, onError);
  return {
    getState,
    getIsActive: () => isActive,
    getSaveFactCachePromise,
    subscribe,
    unsubscribe,
    add: fact => {
      if (isActive) {
        if (factId >= Number.MAX_SAFE_INTEGER) throw new Error(`factId is too big: ${factId}`);
        fact.id = factId + 1;
        factId++;
        factLogger.add(encodeFact(fact));
        setState(applyFact(getState(), fact));
      }
    },
    save: () => {
      isActive && factLogger.save();
    },
    split: () => {
      if (isActive) {
        factLogger.split();
        saveFactCache();
      }
    },
    end: () => {
      if (isActive) {
        isActive = !1;
        factLogger.end();
        saveFactCache();
      }
    }
  };
};
const tryLoadFactInfo = async (factInfo, {applyFact, decodeFact, pathFactDirectory, nameFactLogFile, nameFactCacheFile}) => {
  const factLogFileList = [];
  const factCacheFileList = [];
  const {error} = await catchAsync(async () => (await getDirectorySubInfoList(pathFactDirectory)).forEach(({path, name}) => {
    name.startsWith(nameFactLogFile) && REGEXP_LOG_FILE.test(name) && factLogFileList.push({
      fileId: parseInt(REGEXP_LOG_FILE.exec(name)[1]),
      path,
      name
    });
    name.startsWith(nameFactCacheFile) && REGEXP_CACHE_FILE.test(name) && factCacheFileList.push({
      fileId: parseInt(REGEXP_CACHE_FILE.exec(name)[1]),
      path,
      name
    });
  }));
  factInfo = await tryLoadFactInfoFromCache(factInfo, {
    factCacheFileList
  });
  return await tryLoadFactInfoFromLog(factInfo, {
    factLogFileList,
    decodeFact,
    applyFact
  });
};
const REGEXP_LOG_FILE = /\.(\d+)\.log$/;
const REGEXP_CACHE_FILE = /\.(\d+)\.json$/;
const tryLoadFactInfoFromCache = async (factInfo, {factCacheFileList}) => {
  factCacheFileList.sort((a, b) => b.fileId - a.fileId);
  for (const {path, name} of factCacheFileList) {
    const {factId, factState} = tryParseJSONObject(await readFileAsync(path));
    if (factId) return {
      ...factInfo,
      factId,
      factState,
      factCacheFile: path
    };
  }
  return factInfo;
};
const tryLoadFactInfoFromLog = async (factInfo, {factLogFileList, decodeFact, applyFact}) => {
  let {factId, factState, factLogFile = ""} = factInfo;
  const maxFactLogId = factLogFileList.reduce((o, {fileId}) => fileId <= factId + 1 ? Math.max(o, fileId) : o, 0);
  factLogFileList = factLogFileList.filter(({fileId}) => fileId >= maxFactLogId).sort((a, b) => a.fileId - b.fileId);
  for (const {path, name} of factLogFileList) await createReadlineFromFileAsync(path, logText => {
    const fact = logText && decodeFact(logText);
    if (fact && !(fact.id <= factId)) {
      if (fact.id !== factId + 1) throw new Error(`invalid factId: ${fact.id}, should be: ${factId + 1}. file: ${name}`);
      factState = applyFact(factState, fact);
      factId = fact.id;
      factLogFile = path;
    }
  });
  return {
    ...factInfo,
    factId,
    factState,
    factLogFile
  };
};
const tryDeleteExtraCache = async ({pathFactDirectory, nameFactCacheFile = "factCache", keepFactId = 1 / 0, keepFileCount = 2}) => {
  const factCacheFileList = (await getDirectorySubInfoList(pathFactDirectory)).map(({path, name}) => {
    const fileId = name.startsWith(nameFactCacheFile) && REGEXP_CACHE_FILE.test(name) && parseInt(REGEXP_CACHE_FILE.exec(name)[1]);
    return Number.isInteger(fileId) && {
      fileId,
      path,
      name
    };
  }).filter(Boolean);
  if (!factCacheFileList.length) return;
  factCacheFileList.sort((a, b) => b.fileId - a.fileId);
  let skippedFile = 0;
  for (const {fileId, path, name} of factCacheFileList) {
    if (fileId >= keepFactId) continue;
    if (skippedFile < keepFileCount) {
      skippedFile += 1;
      continue;
    }
    const {error} = await catchAsync(unlinkAsync, path);
  }
};
export { INITIAL_FACT_INFO, createFactDatabase, tryLoadFactInfo, tryDeleteExtraCache };