import { request as httpRequest } from "http";
import { request as httpsRequest } from "https";
import { createGunzip } from "zlib";
import { withRetryAsync } from "../common/function";
import { receiveBufferAsync, toArrayBuffer } from "./data/Buffer";
const requestAsync = (url, option, body) => new Promise((resolve, reject) => {
  const urlObject = (url => url instanceof URL ? url : new URL(url))(url);
  const request = ("https:" === urlObject.protocol ? httpsRequest : httpRequest)(urlObject, option, resolve);
  const endWithError = error => {
    request.destroy();
    error.urlObject = urlObject;
    error.option = option;
    reject(error);
  };
  request.on("timeout", () => endWithError(new Error("NETWORK_TIMEOUT")));
  request.on("error", endWithError);
  request.end(body);
});
const ping = async (url, {method = "GET", headers, body, wait = 5e3, maxRetry = 0} = {}) => withRetryAsync(async () => {
  (await requestAsync(url, {
    method,
    headers,
    timeout: wait
  }, body)).destroy();
}, maxRetry, wait);
const fetchLikeRequest = async (url, {method = "GET", headers: requestHeaders, body, timeout = 1e4} = {}) => {
  const option = {
    method,
    headers: {
      "accept-encoding": "gzip",
      ...requestHeaders
    },
    timeout
  };
  const response = await requestAsync(url, option, body);
  const status = response.statusCode;
  return {
    status,
    ok: status >= 200 && status < 300,
    headers: response.headers,
    ...wrapResponse(response, timeout)
  };
};
const wrapResponse = (response, timeout) => {
  let isKeep;
  let isDropped;
  process.nextTick(() => {
    if (!isKeep) {
      response.destroy();
      isDropped = !0;
    }
  });
  const stream = () => {
    if (isKeep) throw new Error("PAYLOAD_ALREADY_USED");
    if (isDropped) throw new Error("PAYLOAD_ALREADY_DROPPED");
    isKeep = !0;
    const timeoutToken = timeout && setTimeout(() => {
      response.destroy();
      response.emit("error", new Error("PAYLOAD_TIMEOUT"));
    }, timeout);
    response.on("end", () => timeoutToken && clearTimeout(timeoutToken));
    return "gzip" === response.headers["content-encoding"] ? response.pipe(createGunzip()) : response;
  };
  const buffer = async () => receiveBufferAsync(stream());
  const text = () => buffer().then(toText);
  return {
    stream,
    buffer,
    arrayBuffer: () => buffer().then(toArrayBuffer),
    text,
    json: () => text().then(parseJSON)
  };
};
const toText = buffer => buffer.toString();
const parseJSON = text => JSON.parse(text);
export { requestAsync, ping, fetchLikeRequest };