import { createCacheMap } from "../../../common/data/CacheMap";
import { getMIMETypeFromFileName } from "../../../common/module/MIME";
import { readFileAsync, createReadStream } from "../../file/function";
import { getPathStat } from "../../file/File";
import { getWeakEntityTagByStat } from "../../module/EntityTag";
import { responderSendBuffer, responderSendBufferRange, responderSendBufferCompress, responderSendStream, responderSendStreamRange } from "./Send";
const createDefaultCacheMap = () => createCacheMap({
  valueSizeSumMax: 33554432
});
const createResponderBufferCache = ({getBufferData, sizeSingleMax = 524288, expireTime = 6e4, isEnableGzip = !1, serveCacheMap = createDefaultCacheMap()}) => {
  const responderSendCacheBuffer = isEnableGzip ? responderSendBufferCompress : responderSendBuffer;
  return async (store, cacheKey) => {
    let bufferData = serveCacheMap.get(cacheKey);
    bufferData || (bufferData = await getBufferData(store, cacheKey)).length <= sizeSingleMax && serveCacheMap.set(cacheKey, bufferData, bufferData.length, Date.now() + expireTime);
    return responderSendCacheBuffer(store, bufferData);
  };
};
const createResponderServeStatic = ({sizeSingleMax = 524288, expireTime = 6e4, isEnableGzip = !1, isEnableRange = !0, serveCacheMap = createDefaultCacheMap()}) => {
  const serveCache = async (store, filePath, encoding, range) => {
    const bufferData = serveCacheMap.get(filePath);
    if (!bufferData) return !1;
    encoding && store.response.setHeader("content-encoding", encoding);
    if (range) {
      range[1] === 1 / 0 && (range[1] = bufferData.length - 1);
      await responderSendBufferRange(store, bufferData, range);
    } else await responderSendBuffer(store, bufferData);
    return !0;
  };
  const serve = async (store, filePath, type, encoding, range) => {
    const stat = await getPathStat(filePath);
    if (!stat.isFile() || !stat.size) return !1;
    const length = stat.size;
    const entityTag = getWeakEntityTagByStat(stat);
    encoding && store.response.setHeader("content-encoding", encoding);
    if (range) {
      range[1] === 1 / 0 && (range[1] = length - 1);
      await responderSendStreamRange(store, {
        streamRange: createReadStream(filePath, {
          start: range[0],
          end: range[1]
        }),
        length,
        type,
        entityTag
      }, range);
    } else if (length > sizeSingleMax) await responderSendStream(store, {
      stream: createReadStream(filePath),
      length,
      type,
      entityTag
    }); else {
      const bufferData = {
        buffer: await readFileAsync(filePath),
        length,
        type,
        entityTag
      };
      serveCacheMap.set(filePath, bufferData, length, Date.now() + expireTime);
      await responderSendBuffer(store, bufferData);
    }
    return !0;
  };
  return async (store, filePath) => {
    const acceptGzip = isEnableGzip && REGEXP_ENCODING_GZIP.test(store.request.headers["accept-encoding"]);
    const range = isEnableRange && parseRangeHeader(store.request.headers.range);
    isEnableRange && !acceptGzip && store.response.setHeader("accept-ranges", "bytes");
    if (acceptGzip && await serveCache(store, filePath + ".gz", "gzip")) return;
    if (await serveCache(store, filePath, void 0, range)) return;
    const type = getMIMETypeFromFileName(filePath);
    if (!(acceptGzip && await serve(store, filePath + ".gz", type, "gzip") || await serve(store, filePath, type, void 0, range))) throw new Error(`miss file: ${filePath}`);
  };
};
const REGEXP_ENCODING_GZIP = /gzip/i;
const parseRangeHeader = rangeString => {
  const result = REGEXP_HEADER_RANGE.exec(rangeString);
  if (!result) return;
  const [, startString, endString] = result;
  const start = parseInt(startString);
  const end = endString ? parseInt(endString) : 1 / 0;
  return start < end ? [ start, end ] : void 0;
};
const REGEXP_HEADER_RANGE = /bytes=(\d+)-(\d+)?$/i;
export { createDefaultCacheMap, createResponderBufferCache, createResponderServeStatic };