import { constants } from "crypto";
import { createServer as createHttpServer } from "http";
import { createServer as createHttpsServer } from "https";
import { clock } from "../../common/time";
import { createCacheMap } from "../../common/data/CacheMap";
import { createStateStoreLite } from "../../common/immutable/StateStore";
import { responderEnd } from "./Responder/Common";
const DEFAULT_HTTPS_OPTION = {
  protocol: "https:",
  hostname: "localhost",
  port: 443,
  isSecure: !0,
  secureOptions: constants.SSL_OP_NO_SSLv3 | constants.SSL_OP_NO_SSLv2
};
const DEFAULT_HTTP_OPTION = {
  protocol: "http:",
  hostname: "localhost",
  port: 80,
  isSecure: !1
};
const VALID_SERVER_PROTOCOL_SET = new Set([ DEFAULT_HTTPS_OPTION.protocol, DEFAULT_HTTP_OPTION.protocol ]);
const createServer = ({protocol, ...option}) => {
  if (!VALID_SERVER_PROTOCOL_SET.has(protocol)) throw new Error(`invalid protocol: ${protocol}`);
  (option = {
    ..."https:" === protocol ? DEFAULT_HTTPS_OPTION : DEFAULT_HTTP_OPTION,
    ...option
  }).baseUrl = `${protocol}//${option.hostname}:${option.port}`;
  const server = option.isSecure ? createHttpsServer(option) : createHttpServer();
  option.isSecure && (server => {
    const sessionCacheMap = createCacheMap({
      valueSizeSumMax: 4096,
      eventHub: null
    });
    server.on("newSession", (sessionId, sessionData, next) => {
      sessionCacheMap.set(sessionId.toString("hex"), sessionData, 1, Date.now() + 6e5);
      next();
    });
    server.on("resumeSession", (sessionId, next) => {
      next(null, sessionCacheMap.get(sessionId.toString("hex")) || null);
    });
  })(server);
  const socketSet = new Set();
  server.on("connection", socket => {
    socketSet.add(socket);
    socket.once("close", () => {
      socketSet.delete(socket);
    });
  });
  return {
    server,
    socketSet,
    option,
    start: async () => !server.listening && new Promise((resolve, reject) => {
      server.on("error", reject);
      server.listen(option.port, option.hostname, () => {
        server.off("error", reject);
        resolve();
      });
    }),
    stop: async ({isForce = !1} = {}) => server.listening && new Promise(resolve => {
      server.close(() => resolve());
      if (isForce) for (const socket of socketSet) socket.destroy();
    })
  };
};
const DEFAULT_RESPONDER_ERROR = (store, error) => store.setState({
  error
});
const DEFAULT_RESPONDER_END = responderEnd;
const createRequestListener = ({responderList = [], responderError = DEFAULT_RESPONDER_ERROR, responderEnd = DEFAULT_RESPONDER_END}) => async (request, response) => {
  const stateStore = createStateStoreLite({
    time: clock(),
    error: null
  });
  stateStore.request = request;
  stateStore.response = response;
  try {
    for (const responder of responderList) await responder(stateStore);
  } catch (error) {
    await responderError(stateStore, error);
  }
  await responderEnd(stateStore);
};
export { createServer, createRequestListener };