import { createEventEmitter } from "../../../common/module/Event";
import { FRAME_CONFIG, OPCODE_TYPE, WEBSOCKET_EVENT } from "./function";
import { createFrameSenderStore, encodeFrame, encodeCloseFrame, encodePingFrame, encodePongFrame, sendEncodedFrame } from "./frameSender";
import { createFrameReceiverStore, listenAndReceiveFrame } from "./frameReceiver";
const NULL_ERROR_LISTENER = error => {};
const DEFAULT_BUFFER = Buffer.from("Dr");
const createWebSocket = ({socket, frameLengthLimit = 8388608, isMask = !1, shouldPing = !1}) => {
  const eventEmitter = createEventEmitter();
  const frameSenderStore = createFrameSenderStore(frameLengthLimit);
  const frameReceiverStore = createFrameReceiverStore(frameLengthLimit);
  let closeTimeoutToken = null;
  let pingTimeoutToken = null;
  let pongTimeoutToken = null;
  let readyState = 0;
  const doCloseSocket = error => {
    if (3 !== readyState) {
      0 === readyState && socket.writable && socket.write([ "HTTP/1.1 400 Bad Request", "connection: close", "\r\n" ].join("\r\n"));
      readyState = 3;
      closeTimeoutToken && clearTimeout(closeTimeoutToken);
      pingTimeoutToken && clearTimeout(pingTimeoutToken);
      pongTimeoutToken && clearTimeout(pongTimeoutToken);
      closeTimeoutToken = null;
      pingTimeoutToken = null;
      pongTimeoutToken = null;
      frameSenderStore.dispose();
      frameReceiverStore.dispose();
      socket.off("error", close);
      socket.off("end", close);
      socket.on("error", NULL_ERROR_LISTENER);
      socket.destroyed || socket.destroy();
      eventEmitter.emit(WEBSOCKET_EVENT.CLOSE);
    }
  };
  let frameDataType = null;
  let frameBufferList = [];
  let frameBufferLength = 0;
  const onReceiveFrame = frame => {
    switch (frame.dataType) {
     case OPCODE_TYPE.CLOSE:
      {
        const code = frame.dataBufferLength >= 2 && frame.dataBuffer.readUInt16BE(0, !0) || 1e3;
        const reason = frame.dataBufferLength >= 3 && frame.dataBuffer.slice(2, frame.dataBufferLength).toString() || "";
        return close(code, reason);
      }
     case OPCODE_TYPE.PING:
      return sendPong(frame.dataBuffer);
     case OPCODE_TYPE.PONG:
      return receivePong();
    }
    const completeFrameData = (frame => {
      frameDataType || (frameDataType = frame.dataType);
      frameBufferList.push(frame.dataBuffer);
      if ((frameBufferLength += frame.dataBuffer.length) > frameLengthLimit) throw new Error(`frameBufferList length ${frameBufferLength} exceeds limit: ${frameLengthLimit}`);
      if (!frame.isFIN) return null;
      const dataType = frameDataType;
      const dataBuffer = 1 === frameBufferList.length ? frameBufferList[0] : Buffer.concat(frameBufferList, frameBufferLength);
      frameDataType = null;
      frameBufferList = [];
      frameBufferLength = 0;
      return {
        dataType,
        dataBuffer
      };
    })(frame);
    completeFrameData && eventEmitter.emit(WEBSOCKET_EVENT.FRAME, completeFrameData);
    shouldPing && setNextPing();
  };
  const close = (code = 1e3, reason = "") => {
    if (3 !== readyState) {
      if (2 === readyState) return doCloseSocket();
      if (1 !== readyState && 2 !== readyState) throw new Error(`error readyState = ${readyState}`);
      readyState = 2;
      closeTimeoutToken = setTimeout(doCloseSocket, 5e3);
      encodeCloseFrame(frameSenderStore, code, reason, isMask);
      1e3 === code ? sendEncodedFrame(frameSenderStore, socket).catch(doCloseSocket) : sendEncodedFrame(frameSenderStore, socket).then(doCloseSocket, doCloseSocket);
    }
  };
  const setNextPing = () => {
    pingTimeoutToken && clearTimeout(pingTimeoutToken);
    pingTimeoutToken = setTimeout(sendPing, 6e4);
  };
  const setNextPong = () => {
    pongTimeoutToken && clearTimeout(pongTimeoutToken);
    pongTimeoutToken = setTimeout(() => close(1006, "pong timeout"), 6e4);
  };
  const sendPing = (dataBuffer = DEFAULT_BUFFER) => {
    if (1 === readyState) {
      setNextPong();
      encodePingFrame(frameSenderStore, dataBuffer, isMask);
      return sendEncodedFrame(frameSenderStore, socket);
    }
  };
  const sendPong = (dataBuffer = DEFAULT_BUFFER) => {
    if (1 === readyState) {
      encodePongFrame(frameSenderStore, dataBuffer, isMask);
      return sendEncodedFrame(frameSenderStore, socket);
    }
  };
  const receivePong = () => {
    pongTimeoutToken && clearTimeout(pongTimeoutToken);
    pongTimeoutToken = null;
    if (shouldPing) setNextPing(); else {
      pingTimeoutToken && clearTimeout(pingTimeoutToken);
      pingTimeoutToken = null;
    }
  };
  return {
    CONNECTING: 0,
    OPEN: 1,
    CLOSING: 2,
    CLOSED: 3,
    ...eventEmitter,
    socket,
    frameLengthLimit,
    isMask,
    getReadyState: () => readyState,
    isClosed: () => 3 === readyState || !socket || socket.destroyed,
    doCloseSocket,
    open: () => {
      socket.on("error", close);
      socket.on("end", close);
      listenAndReceiveFrame(frameReceiverStore, socket, onReceiveFrame, error => close(1006, "Frame Error"));
      readyState = 1;
      eventEmitter.emit(WEBSOCKET_EVENT.OPEN);
    },
    close,
    sendText: text => {
      if (1 !== readyState) throw new Error(`not open yet: readyState = ${readyState}`);
      encodeFrame(frameSenderStore, FRAME_CONFIG.COMPLETE, OPCODE_TYPE.TEXT, Buffer.from(text), isMask);
      return sendEncodedFrame(frameSenderStore, socket);
    },
    sendBuffer: buffer => {
      if (1 !== readyState) throw new Error(`not open yet: readyState = ${readyState}`);
      encodeFrame(frameSenderStore, FRAME_CONFIG.COMPLETE, OPCODE_TYPE.BINARY, buffer, isMask);
      return sendEncodedFrame(frameSenderStore, socket);
    },
    setNextPing,
    setNextPong,
    sendPing,
    sendPong
  };
};
export { createWebSocket };