import { BUFFER_MAX_LENGTH, applyMaskQuadletBufferInPlace } from "./function";
const NULL_ERROR = error => {};
const createFrameReceiverStore = frameLengthLimit => {
  let promiseTail = Promise.resolve("HEAD");
  let doClearSocketListener = null;
  return {
    dispose: () => {
      promiseTail.then(NULL_ERROR, NULL_ERROR);
      promiseTail = null;
      doClearSocketListener && doClearSocketListener();
      doClearSocketListener = null;
    },
    queuePromise: (resolve, reject) => promiseTail = promiseTail.then(resolve, reject),
    setClearSocketListener: nextDoClearSocketListener => {
      doClearSocketListener = nextDoClearSocketListener;
    },
    frameDecoder: createFrameDecoder(frameLengthLimit),
    frameLengthLimit
  };
};
const listenAndReceiveFrame = (frameReceiverStore, socket, onFrame, onError = frameReceiverStore.dispose) => {
  let receiveResolve = null;
  let receiveReject = null;
  const reset = () => {
    receiveReject && socket.off("error", receiveReject);
    receiveResolve = null;
    receiveReject = null;
    frameReceiverStore.frameDecoder.resetDecode();
  };
  const onPromiseReject = error => {
    reset();
    onError(error);
  };
  const promiseReceive = () => {
    const receivePromise = new Promise((resolve, reject) => {
      receiveResolve = resolve;
      receiveReject = reject;
    });
    socket.on("error", receiveReject);
    frameReceiverStore.queuePromise(() => receivePromise.then(onFrame), onPromiseReject);
  };
  const onSocketData = chunk => {
    frameReceiverStore.frameDecoder.pushBufferFrag(chunk);
    for (;;) {
      null === receiveResolve && promiseReceive();
      if (!frameReceiverStore.frameDecoder.decode()) break;
      const frame = frameReceiverStore.frameDecoder.tryGetDecodedFrame();
      if (void 0 !== frame) {
        receiveResolve(frame);
        reset();
      }
    }
  };
  socket.on("data", onSocketData);
  frameReceiverStore.setClearSocketListener(() => {
    socket.off("data", onSocketData);
    reset();
  });
};
const createFrameDecoder = frameLengthLimit => {
  const {pushBufferFrag, tryShiftMergedBuffer} = createBufferPool();
  let mergedBuffer = null;
  let stage = 0;
  let decodedIsMask = !1;
  let decodedMaskQuadletBuffer = null;
  let decodedIsFIN = !1;
  let decodedDataType = null;
  let decodedDataBuffer = null;
  let decodedDataBufferLength = 0;
  return {
    pushBufferFrag,
    decode: () => {
      switch (stage) {
       case 0:
        if (mergedBuffer = tryShiftMergedBuffer(2)) {
          const initialQuadlet = mergedBuffer.readUInt16BE(0, !0);
          const initialLength = 127 & initialQuadlet;
          decodedIsMask = 0 != (128 & initialQuadlet);
          decodedIsFIN = 8 == (initialQuadlet >>> 12 & 8);
          decodedDataType = initialQuadlet >>> 8 & 15;
          if (0 === initialLength) {
            decodedDataBufferLength = 0;
            stage = decodedIsMask ? 3 : 5;
          } else if (initialLength <= 125) {
            if ((decodedDataBufferLength = initialLength) > frameLengthLimit) throw new Error(`dataBuffer length ${decodedDataBufferLength} exceeds limit: ${frameLengthLimit}`);
            stage = decodedIsMask ? 3 : 4;
          } else stage = 126 === initialLength ? 1 : 2;
          return !0;
        }
        break;
       case 1:
        if (mergedBuffer = tryShiftMergedBuffer(2)) {
          if ((decodedDataBufferLength = mergedBuffer.readUInt16BE(0, !0)) > frameLengthLimit) throw new Error(`dataBuffer length ${decodedDataBufferLength} exceeds limit: ${frameLengthLimit}`);
          stage = decodedIsMask ? 3 : 4;
          return !0;
        }
        break;
       case 2:
        if (mergedBuffer = tryShiftMergedBuffer(8)) {
          if ((decodedDataBufferLength = 4294967296 * mergedBuffer.readUInt32BE(0, !0) + mergedBuffer.readUInt32BE(4, !0)) > BUFFER_MAX_LENGTH) throw new Error("decodedDataBufferLength too big");
          if (decodedDataBufferLength > frameLengthLimit) throw new Error(`dataBuffer length ${decodedDataBufferLength} exceeds limit: ${frameLengthLimit}`);
          stage = decodedIsMask ? 3 : 4;
          return !0;
        }
        break;
       case 3:
        if (mergedBuffer = tryShiftMergedBuffer(4)) {
          decodedMaskQuadletBuffer = mergedBuffer;
          stage = decodedDataBufferLength ? 4 : 5;
          return !0;
        }
        break;
       case 4:
        if (mergedBuffer = tryShiftMergedBuffer(decodedDataBufferLength)) {
          decodedDataBuffer = mergedBuffer;
          decodedIsMask && applyMaskQuadletBufferInPlace(decodedDataBuffer, decodedMaskQuadletBuffer);
          stage = 5;
          return !0;
        }
      }
      return !1;
    },
    resetDecode: () => {
      stage = 0;
      decodedIsMask = !1;
      decodedMaskQuadletBuffer = null;
      decodedIsFIN = !1;
      decodedDataType = null;
      decodedDataBuffer = null;
      decodedDataBufferLength = 0;
    },
    tryGetDecodedFrame: () => 5 !== stage ? void 0 : {
      isFIN: decodedIsFIN,
      dataType: decodedDataType,
      dataBuffer: decodedDataBuffer,
      dataBufferLength: decodedDataBufferLength
    }
  };
};
const createBufferPool = () => {
  const pool = [];
  let poolSumLength = 0;
  return {
    pushBufferFrag: bufferFrag => {
      pool.push(bufferFrag);
      poolSumLength += bufferFrag.length;
    },
    tryShiftMergedBuffer: length => {
      if (!(poolSumLength < length)) {
        poolSumLength -= length;
        if (length === pool[0].length) return pool.shift();
        if (length < pool[0].length) {
          const buffer = pool[0].slice(0, length);
          pool[0] = pool[0].slice(length);
          return buffer;
        }
        {
          const buffer = Buffer.allocUnsafe(length);
          let offset = 0;
          for (;length > 0; ) {
            const frag = pool[0];
            const fragLength = frag.length;
            if (length >= fragLength) {
              frag.copy(buffer, offset);
              pool.shift();
              offset += fragLength;
              length -= fragLength;
            } else {
              frag.copy(buffer, offset, 0, length);
              pool[0] = frag.slice(length);
              length = 0;
            }
          }
          return buffer;
        }
      }
    }
  };
};
export { createFrameReceiverStore, listenAndReceiveFrame };