import { randomBytes } from "crypto";
import { FRAME_CONFIG, OPCODE_TYPE, BUFFER_MAX_LENGTH, applyMaskQuadletBufferInPlace } from "./function";
const DEFAULT_MASK_QUADLET_BUFFER = Buffer.alloc(4);
const NULL_ERROR = error => {};
const createFrameSenderStore = frameLengthLimit => {
  let promiseTail = Promise.resolve("HEAD");
  return {
    dispose: () => {
      promiseTail.then(NULL_ERROR, NULL_ERROR);
      promiseTail = null;
    },
    queuePromise: (resolve, reject) => promiseTail = promiseTail.then(resolve, reject),
    encodedFrameHeaderBuffer: null,
    encodedFrameDataBuffer: null,
    frameLengthLimit
  };
};
const encodeFrame = (frameSenderStore, frameTypeConfig, dataType, dataBuffer, isMask) => {
  const [quadbitFIN, quadbitOpcodeMask] = frameTypeConfig;
  const {length} = dataBuffer;
  const initialOctet = quadbitFIN << 4 | dataType & quadbitOpcodeMask;
  let maskLengthOctet, extendLengthOctetCount;
  if (length <= 125) {
    maskLengthOctet = length;
    extendLengthOctetCount = 0;
  } else if (length <= 65535) {
    maskLengthOctet = 126;
    extendLengthOctetCount = 2;
  } else {
    if (!(length <= BUFFER_MAX_LENGTH)) throw new Error("dataBuffer length too big");
    maskLengthOctet = 127;
    extendLengthOctetCount = 8;
  }
  if (length > frameSenderStore.frameLengthLimit) throw new Error(`dataBuffer length ${length} exceeds limit: ${frameSenderStore.frameLengthLimit}`);
  const maskOctetCount = isMask ? 4 : 0;
  const maskQuadletBuffer = isMask && length ? randomBytes(4) : DEFAULT_MASK_QUADLET_BUFFER;
  isMask && (maskLengthOctet |= 128);
  frameSenderStore.encodedFrameHeaderBuffer = Buffer.allocUnsafe(2 + extendLengthOctetCount + maskOctetCount);
  frameSenderStore.encodedFrameHeaderBuffer.writeUInt16BE(initialOctet << 8 | maskLengthOctet, 0, !0);
  2 === extendLengthOctetCount && frameSenderStore.encodedFrameHeaderBuffer.writeUInt16BE(length, 2, !0);
  8 === extendLengthOctetCount && frameSenderStore.encodedFrameHeaderBuffer.writeUInt32BE(0, 2, !0);
  8 === extendLengthOctetCount && frameSenderStore.encodedFrameHeaderBuffer.writeUInt32BE(length, 6, !0);
  isMask && maskQuadletBuffer.copy(frameSenderStore.encodedFrameHeaderBuffer, 2 + extendLengthOctetCount);
  frameSenderStore.encodedFrameDataBuffer = dataBuffer;
  isMask && length && applyMaskQuadletBufferInPlace(frameSenderStore.encodedFrameDataBuffer, maskQuadletBuffer);
};
const encodeCloseFrame = (frameSenderStore, code = 1e3, reason = "", isMask) => {
  const stringLength = Buffer.byteLength(reason);
  const dataBuffer = Buffer.allocUnsafe(2 + stringLength);
  dataBuffer.writeUInt16BE(code, 0, !0);
  dataBuffer.write(reason, 2, stringLength);
  encodeFrame(frameSenderStore, FRAME_CONFIG.COMPLETE, OPCODE_TYPE.CLOSE, dataBuffer, isMask);
};
const encodePingFrame = (frameSenderStore, dataBuffer, isMask) => encodeFrame(frameSenderStore, FRAME_CONFIG.COMPLETE, OPCODE_TYPE.PING, dataBuffer, isMask);
const encodePongFrame = (frameSenderStore, dataBuffer, isMask) => encodeFrame(frameSenderStore, FRAME_CONFIG.COMPLETE, OPCODE_TYPE.PONG, dataBuffer, isMask);
const sendEncodedFrame = (frameSenderStore, socket) => {
  const frameHeaderBuffer = frameSenderStore.encodedFrameHeaderBuffer;
  const frameDataBuffer = frameSenderStore.encodedFrameDataBuffer;
  frameSenderStore.encodedFrameHeaderBuffer = null;
  frameSenderStore.encodedFrameDataBuffer = null;
  return frameSenderStore.queuePromise(() => new Promise((resolve, reject) => {
    const onDataSend = () => {
      socket.off("error", reject);
      resolve();
    };
    socket.on("error", reject);
    if (0 === frameDataBuffer.length) socket.write(frameHeaderBuffer, onDataSend); else {
      socket.write(frameHeaderBuffer);
      socket.write(frameDataBuffer, onDataSend);
    }
  }));
};
export { createFrameSenderStore, encodeFrame, encodeCloseFrame, encodePingFrame, encodePongFrame, sendEncodedFrame };